#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2009 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________


import gc
import os
import pstats
import sys
import traceback

from optparse import OptionParser, OptionGroup

try:
    import cProfile as profile
except ImportError:
    import profile

from coopr.opt.base import SolverFactory
from coopr.opt.parallel import SolverManagerFactory
from coopr.pysp.ef import *
from coopr.pysp.solutionwriter import ISolutionWriterExtension
from pyutilib.component.core import ExtensionPoint
from pyutilib.services import TempfileManager

#
# utility method to construct an option parser for ef writer arguments
#

def construct_ef_writer_options_parser(usage_string):

   solver_list = SolverFactory.services()
   solver_list = sorted( filter(lambda x: '_' != x[0], solver_list) )
   solver_help = \
   "Specify the solver with which to solve scenario sub-problems.  The "      \
   "following solver types are currently supported: %s; Default: cplex"
   solver_help %= ', '.join( solver_list )

   parser = OptionParser()
   parser.usage=usage_string

   inputOpts        = OptionGroup( parser, 'Input Options' )
   scenarioTreeOpts = OptionGroup( parser, 'Scenario Tree Options' )
   efOpts           = OptionGroup( parser, 'EF Options' )
   solverOpts       = OptionGroup( parser, 'Solver Options' )
   outputOpts       = OptionGroup( parser, 'Output Options' )
   otherOpts        = OptionGroup( parser, 'Other Options' )
   parser.add_option_group( inputOpts )
   parser.add_option_group( scenarioTreeOpts )
   parser.add_option_group( efOpts )
   parser.add_option_group( solverOpts )
   parser.add_option_group( outputOpts )
   parser.add_option_group( otherOpts )

   inputOpts.add_option('-i','--instance-directory',
     help='The directory in which all instance (reference and scenario) definitions are stored. Default is ".".',
     action='store',
     dest='instance_directory',
     type='string',
     default='.')
   inputOpts.add_option('-m','--model-directory',
     help='The directory in which all model (reference and scenario) definitions are stored. Default is ".".',
     action='store',
     dest='model_directory',
     type='string',
     default='.')

   scenarioTreeOpts.add_option('--scenario-tree-seed',
     help="The random seed associated with manipulation operations on the scenario tree (e.g., down-sampling). Default is 0, indicating unassigned.",
     action="store",
     dest="scenario_tree_random_seed",
     type="int",
     default=None)
   scenarioTreeOpts.add_option('--scenario-tree-downsample-fraction',
     help="The proportion of the scenarios in the scenario tree that are actually used. Specific scenarios are selected at random. Default is 1.0, indicating no down-sampling.",
     action="store",
     dest="scenario_tree_downsample_fraction",
     type="float",
     default=1.0)

   efOpts.add_option('--cvar-weight',
     help='The weight associated with the CVaR term in the risk-weighted objective formulation. Default is 1.0. If the weight is 0, then *only* a non-weighted CVaR cost will appear in the EF objective - the expected cost component will be dropped.',
     action='store',
     dest='cvar_weight',
     type='float',
     default=1.0)
   efOpts.add_option('--generate-weighted-cvar',
     help='Add a weighted CVaR term to the primary objective',
     action='store_true',
     dest='generate_weighted_cvar',
     default=False)
   efOpts.add_option('--risk-alpha',
     help='The probability threshold associated with cvar (or any future) risk-oriented performance metrics. Default is 0.95.',
     action='store',
     dest='risk_alpha',
     type='float',
     default=0.95)
   efOpts.add_option("--flatten-expressions", "--linearize-expressions",
     help="EXPERIMENTAL: An option intended for use on linear or mixed-integer models " \
          "in which expression trees in a model (constraints or objectives) are compacted " \
          "into a more memory-efficient and concise form. The trees themselves are eliminated. ",
     action="store_true",
     dest="linearize_expressions",
     default=False)

   solverOpts.add_option('--mipgap',
     help='Specifies the mipgap for the EF solve.',
     action='store',
     dest='mipgap',
     type='float',
     default=None)
   solverOpts.add_option('--solve',
     help='Following write of the extensive form model, solve it.',
     action='store_true',
     dest='solve_ef',
     default=False)
   solverOpts.add_option('--solver',
     help=solver_help,
     action='store',
     dest='solver_type',
     type='string',
     default='cplex')
   solverOpts.add_option('--solver-manager',
     help='The type of solver manager used to coordinate scenario sub-problem solves. Default is serial.',
     action='store',
     dest='solver_manager_type',
     type='string',
     default='serial')
   solverOpts.add_option('--solver-options',
     help='Solver options for the extension form problem.',
     action='append',
     dest='solver_options',
     type='string',
     default=[])

   outputOpts.add_option('--output-file',
     help='Specify the name of the extensive form output file',
     action='store',
     dest='output_file',
     type='string',
     default='efout.lp')
   outputOpts.add_option('--output-solver-log',
     help='Output solver log during the extensive form solve.',
     action='store_true',
     dest='output_solver_log',
     default=False)
   outputOpts.add_option('--solution-writer',
     help='The plugin invoked to write the scenario tree solution. Defaults to the empty list.',
     action='append',
     dest='solution_writer',
     type='string',
     default = [])
   outputOpts.add_option('--verbose',
     help='Generate verbose output, beyond the usual status output. Default is False.',
     action='store_true',
     dest='verbose',
     default=False)

   otherOpts.add_option('--disable-gc',
     help='Disable the python garbage collecter. Default is False.',
     action='store_true',
     dest='disable_gc',
     default=False)
   otherOpts.add_option('-k','--keep-solver-files',
     help='Retain temporary input and output files for solve.',
     action='store_true',
     dest='keep_solver_files',
     default=False)
   otherOpts.add_option('--profile',
     help='Enable profiling of Python code.  The value of this option is the number of functions that are summarized.',
     action='store',
     dest='profile',
     default=0)
   otherOpts.add_option('--traceback',
     help='When an exception is thrown, show the entire call stack. Ignored if profiling is enabled. Default is False.',
     action='store_true',
     dest='traceback',
     default=False)

   return parser


def run_ef_writer(options, args):

   # if the user enabled the addition of the weighted cvar term to the objective,
   # then validate the associated parameters.
   generate_weighted_cvar = False
   cvar_weight = None
   risk_alpha = None

   if options.generate_weighted_cvar is True:

      generate_weighted_cvar = True
      cvar_weight = options.cvar_weight
      risk_alpha = options.risk_alpha

   # validate the solution writer plugin exists, to avoid a lot of wasted work.
   for solution_writer_name in options.solution_writer:
      print "Trying to import solution writer="+solution_writer_name
      __import__(solution_writer_name)
      print "Module successfully loaded"

   scenario_tree, binding_instance, scenario_instances = write_ef_from_scratch(os.path.expanduser(options.model_directory),
                                                                               os.path.expanduser(options.instance_directory),
                                                                               os.path.expanduser(options.output_file),
                                                                               options.verbose,
                                                                               options.linearize_expressions,
                                                                               options.scenario_tree_downsample_fraction,
                                                                               options.scenario_tree_random_seed,
                                                                               generate_weighted_cvar, cvar_weight, risk_alpha)

   if (scenario_tree is None) or (binding_instance is None) or (scenario_instances is None):
      raise RuntimeError, "Failed to write extensive form."

   if options.solve_ef is True:

      ef_solver = SolverFactory(options.solver_type)
      if ef_solver is None:
         raise ValueError, "Failed to create solver of type="+options.solver_type+" for use in extensive form solve"
      if len(options.solver_options) > 0:
         print "Initializing ef solver with options="+str(options.solver_options)
         ef_solver.set_options("".join(options.solver_options))
      if options.mipgap is not None:
         if (options.mipgap < 0.0) or (options.mipgap > 1.0):
            raise ValueError, "Value of the mipgap parameter for the EF solve must be on the unit interval; value specified=" + `options.mipgap`
         else:
            ef_solver.mipgap = options.mipgap
      if options.keep_solver_files is True:
         ef_solver.keepFiles = True

      ef_solver_manager = SolverManagerFactory(options.solver_manager_type)
      if ef_solver is None:
         raise ValueError, "Failed to create solver manager of type="+options.solver_type+" for use in extensive form solve"

      # at this point you have a specific solver - communicate solver capabilities
      # to the writer via the instance.
      binding_instance.has_capability = ef_solver.has_capability
      for scenario_name, scenario_instance in scenario_instances.items():
         scenario_instance.has_capability = ef_solver.has_capability

      print "Queuing extensive form solve"
      ef_action_handle = ef_solver_manager.queue(os.path.expanduser(options.output_file), opt=ef_solver, tee=options.output_solver_log)
      print "Waiting for extensive form solve"
      ef_results = ef_solver_manager.wait_for(ef_action_handle)
      load_ef_solution(ef_results, binding_instance, scenario_instances)
      scenario_tree.snapshotSolutionFromInstances(scenario_instances)

      # handle output of solution from the scenario tree.
      print ""
      print "Extensive form solution:"
      scenario_tree.pprintSolution()
      print ""
      print "Extensive form costs:"
      scenario_tree.pprintCosts(scenario_instances)

      solution_writer_plugins = ExtensionPoint(ISolutionWriterExtension)
      for plugin in solution_writer_plugins:
         plugin.write(scenario_tree, "ef")

def main(args=None):

    #
    # Top-level command that executes the extensive form writer.
    # This is segregated from run_ef_writer to enable profiling.
    #

    #
    # Parse command-line options.
    #
    try:
       options_parser = construct_ef_writer_options_parser("runef [options]")
       (options, args) = options_parser.parse_args(args=args)
    except SystemExit:
       # the parser throws a system exit if "-h" is specified - catch
       # it to exit gracefully.
       return

    if options.disable_gc is True:
       gc.disable()
    else:
       gc.enable()

    # if an exception is triggered and traceback is enabled, 'ans' won't
    # have a value and the return statement from this function will flag
    # an error, masking the stack trace that you really want to see.
    ans = None

    if options.profile > 0:
        #
        # Call the main ef writer with profiling.
        #
        tfile = TempfileManager.create_tempfile(suffix=".profile")
        tmp = profile.runctx('run_ef_writer(options,args)',globals(),locals(),tfile)
        p = pstats.Stats(tfile).strip_dirs()
        p.sort_stats('time', 'cum')
        options.profile = eval(options.profile)
        p = p.print_stats(options.profile)
        p.print_callers(options.profile)
        p.print_callees(options.profile)
        p = p.sort_stats('cum','calls')
        p.print_stats(options.profile)
        p.print_callers(options.profile)
        p.print_callees(options.profile)
        p = p.sort_stats('calls')
        p.print_stats(options.profile)
        p.print_callers(options.profile)
        p.print_callees(options.profile)
        TempfileManager.clear_tempfiles()
        ans = [tmp, None]
    else:
        #
        # Call the main EF writer without profiling.
        #
        if options.traceback is True:
           ans = run_ef_writer(options, args)
        else:
           errmsg = None
           try:
              ans = run_ef_writer(options, args)
           except ValueError, err:
              errmsg = 'VALUE ERROR: %s' % err
           except TypeError, err:
              errmsg = 'TYPE ERROR: %s' % err
           except NameError, err:
              errmsg = 'NAME ERROR: %s' % err
           except IOError, err:
              errmsg = 'I/O ERROR: %s' % err
           except RuntimeError, err:
              errmsg = 'RUN-TIME ERROR: %s' % err
           except pyutilib.common.ApplicationError, err:
              errmsg = 'APPLICATION ERROR: %s' % err
           except Exception, err:
              errmsg = 'UNKNOWN ERROR: %s' % err
              traceback.print_exc()

           if errmsg is not None:
              print >>sys.stderr, errmsg

    gc.enable()

    return ans

